#include "rusinluk_dip_incl.h"

Circuit::Circuit() {
	clear(); // init
	state = CIRC_ST_CREATED;
}

Circuit::Circuit(const string &file) {
	load(file);
}

Circuit::~Circuit() {
	clear();
}

const string &Circuit::getFile() const {
	return file;
}

int Circuit::getCwinN() const {
	return cwinN;
}

int Circuit::getGateN() const {
	return gateN;
}

int Circuit::getGateInN() const {
	return gateInN;
}

int Circuit::getLevN() const {
	return levN;
}

void Circuit::load(const string &file) {
	loadOnly(file);
	setEdges();
	gateN = cwins.size();
	setGateInN();
	setLevN();
	state = CIRC_ST_LOADED;
}

void Circuit::makeWins(const Params &pm, MyRand &rnd) {
	checkState(CIRC_ST_LOADED);
	if (pm.cwinN < 1) {
			// don't divide
		cwinN = 1;
		state = CIRC_ST_CWINS;
		return;
	}
	TrieStor<CWin *> neighbors;
	TrieStor<int> workIns, workOuts, workIsect;
	Heap heap(rnd);
	buildEdgeHeap(heap, pm, neighbors, workIns, workOuts, workIsect);
	cwinN = cwins.size();
	while (cwinN > pm.cwinN) {
		CWinEdge *edge = dynCast<Heapable, CWinEdge>(heap.getMin());
		if (edge == NULL) {
			break; // no edges => minimal cwinN is reached
		}
		CWin
			*cwinA = edge->getCWinA(),
			*cwinB = edge->getCWinB();
		cwinA->merge(cwinB, outs, neighbors, workIns, workOuts, workIsect);
			// neighbors are set
		removeEdges(cwinA, heap);
		removeEdges(cwinB, heap);
		cwins[cwinB->getId()] = NULL;
		delete cwinB;
		cwinN--;
		addEdges(cwinA, heap, pm, neighbors, workIns, workOuts, workIsect); // + evaluate
	}
	state = CIRC_ST_CWINS;
}

void Circuit::split(const Params &pm) {
	checkState(CIRC_ST_CWINS);
	string fileW;
	if (pm.cwinN < 1) {
		fileW = file;
		fileSuffix(fileW, FILE_CWIN + toString(0));
		fileCopy(file, fileW);
		state = CIRC_ST_SPLITTED;
		return;
	}
	ofstream ofs;
	int cwinI = -1;
	for (unsigned i = 0; i < cwins.size(); i++) {
		if (cwins[i] == NULL) {
			continue;
		}
		cwinI++;
		fileW = file;
		fileSuffix(fileW, FILE_CWIN + toString(cwinI));
		ofsOpen(ofs, fileW, ios_base::out);
		ofs << (sModel + "_" + toString(cwinI)) << endl;
		ofs << BLIF_INS;
		for (unsigned j = 0; j < cwins[i]->ins.size(); j++) {
			ofs << " " << labels[cwins[i]->ins[j]];
		}
		ofs << endl << BLIF_OUTS;
		for (unsigned j = 0; j < cwins[i]->outs.size(); j++) {
			ofs << " " << labels[cwins[i]->outs[j]];
		}
		ofs << endl;
		Gate *gate = cwins[i]->getGates();
		while (gate != NULL) {
			gate->save(ofs);
			gate = gate->next;
		}
		ofs << BLIF_END << endl;
		ofs.close();
	}
	state = CIRC_ST_SPLITTED;
}

void Circuit::mergeAndSave(const string &fileO) {
	checkState(CIRC_ST_SPLITTED);
	ofstream ofs;
	ofsOpen(ofs, fileO, ios_base::out);
	ofs << sModel << endl;
	ofs << sIns << endl;
	ofs << sOuts << endl;
	vector<Circuit *> circs;
	circs.reserve(cwinN);
	labels.reset(); /* new usage: inter-window signals
		- to avoid after-resynthesis signal collisions */
	vector<string> v;
	string fileW;
	for (int i = 0; i < cwinN; i++) {
		fileW = file;
		fileSuffix(fileW, FILE_OUTPUT + toString(i));
		circs.push_back(new Circuit());
		circs[i]->loadOnly(fileW);
		for (unsigned j = 0; j < circs[i]->cwins.size(); j++) {
			parseNames(circs[i]->cwins[j]->getGates()->stor[0], v); // gate's first line with names
			for (unsigned k = 1; k < v.size(); k++) {
				int hashI = circs[i]->labels.getI(v[k]);
				if (circs[i]->ins.find(hashI) || circs[i]->outs.find(hashI)) {
					labels.add(v[k]);
				}
			}
		}
	}
	TrieStor<string> locNames;
	vector<int> nodeNames;
	string line, nodeName;
	int nodeLastI = 0;
	for (int i = 0; i < cwinN; i++) {
		int n = circs[i]->cwins.size() - circs[i]->outs.size(); // number of local signals to rename
		locNames.reset();
		locNames.reserveData(n);
		locNames.reserveTrie((int)ceil(n * TRIE_STRING_COEF));
		nodeNames.clear();
		nodeNames.reserve(n);
		for (unsigned j = 0; j < circs[i]->cwins.size(); j++) {
			string &names = circs[i]->cwins[j]->getGates()->stor[0];
			parseNames(names, v);
			line = BLIF_NAMES;
			for (unsigned k = 1; k < v.size(); k++) {
				line += " ";
				int hashI = circs[i]->labels.getI(v[k]);
				if (circs[i]->ins.find(hashI) || circs[i]->outs.find(hashI)) {
					line += v[k]; // no change
					continue;
				}
				if (locNames.add(v[k])) {
					do {
						nodeName = CIRC_NODE + toString(++nodeLastI);
					}
						while (labels.find(nodeName));
					nodeNames.push_back(nodeLastI);
				}
				else {
					nodeName = CIRC_NODE + toString(nodeNames[locNames.getI(v[k])]);
				}
				line += nodeName;
			}
			names = line;
			circs[i]->cwins[j]->getGates()->save(ofs);
		}
	}
	ofs << BLIF_END << endl;
	for (unsigned i = 0; i < circs.size(); i++) {
		delete circs[i];
	}
	state = CIRC_ST_MERGED;
}

double Circuit::eval(const Params &pm) const {
	return
		gateN * pm.wgRwinG +
		gateInN * pm.wgRwinGI +
		levN * pm.wgRwinL;
}

void Circuit::printEval() const {
	logRow("Circuit evaluation:");
	logRow("  gates: " + toString(gateN));
	logRow("  gate inputs: " + toString(gateInN));
	logRow("  levels: " + toString(levN));
}

void Circuit::clear() {
	sModel = BLIF_DMODEL;
	sIns = "";
	sOuts = "";
	labels.reset();
	ins.reset();
	outs.reset();
	for (unsigned i = 0; i < cwins.size(); i++) {
		delete cwins[i];
	}
	cwins.clear();
}

void Circuit::checkState(int x) const {
	if (state != x) {
		logErr("Error, bad circuit state: " + toString(state) + ", required " + toString(x));
	}
}

void Circuit::loadOnly(const string &file) {
	clear();
	this->file = file;
	loadBlif();
	setNames();
}

void Circuit::loadBlif() {
	ifstream ifs;
	ifsOpen(ifs, file, ios_base::in);
	string str;
	bool readNext = true;
	while (!readNext || (getBlifLine(ifs, str) && str != BLIF_END)) {
		readNext = true;
		if (str.find(BLIF_MODEL) == 0) {
			sModel = str;
			continue;
		}
		if (str.find(BLIF_INS) == 0) {
			sIns = str;
			continue;
		}
		if (str.find(BLIF_OUTS) == 0) {
			sOuts = str;
			continue;
		}
		Gate *gate = new Gate();
		if (gate->load(ifs, str)) {
			cwins.push_back(new CWin(cwins.size(), gate));
			readNext = false;
		}
		else {
			delete gate;
		}
	}
	if (!cwins.size()) {
		logErr(ERR_BLIF, "No gates in file " + file);
	}
}

void Circuit::setNames() {
	vector<string> vIns, vOuts, vGate;
	parseNames(sIns, vIns);
	parseNames(sOuts, vOuts);
	labels.reserveData(cwins.size() + vIns.size());
	labels.reserveTrie((int)ceil((cwins.size() + vIns.size()) * TRIE_STRING_COEF));
	ins.reserveData(vIns.size());
	ins.reserveTrie((int)ceil(vIns.size() * TRIE_INT_COEF));
	outs.reserveData(vOuts.size());
	outs.reserveTrie((int)ceil(vOuts.size() * TRIE_INT_COEF));
	for (unsigned i = 0; i < cwins.size(); i++) {
			// gate outputs
		parseNames(cwins[i]->getGates()->stor[0], vGate);
		if (vGate.size() <= 1) {
			logErr(ERR_BLIF, "No gate output (" + toString(i + 1) + ".) in file " + file);
		}
		if (!labels.add(vGate.back())) {
			logErr(ERR_BLIF, "Duplicit gate output \"" + vGate.back() + "\" in file " + file);
		}
		cwins[i]->outs.add(labels.size() - 1); // hash of the last added label
	}
	for (unsigned i = 1; i < vOuts.size(); i++) {
			// circuit outputs (the driven ones only)
		int hashI = labels.getI(vOuts[i]);
		if (hashI != NO_INDEX) {
			outs.add(hashI);
		}
	}
	for (unsigned i = 1; i < vIns.size(); i++) {
			// circuit inputs
		if (labels.add(vIns[i])) {
			ins.add(labels.size() - 1);
			continue;
		}
		if (labels.getI(vIns[i]) < (int)cwins.size()) {
			logErr(ERR_BLIF, "Gate output \"" + vIns[i] + "\" as a primary input in file " + file);
		}
	}
	for (unsigned i = 0; i < cwins.size(); i++) {
			// gate inputs
		parseNames(cwins[i]->getGates()->stor[0], vGate);
		for (unsigned j = 1; j <= vGate.size() - 2; j++) {
			int hashI = labels.getI(vGate[j]);
			if (hashI == NO_INDEX) {
				logErr(ERR_BLIF, "Unknown gate input \"" + vGate[j] + "\" on gate \"" + vGate.back() + "\" in file " + file);
			}
			cwins[i]->ins.add(hashI);
		}
	}
}

void Circuit::setEdges() {
	edgeN = 0;
	for (unsigned i = 0; i < cwins.size(); i++) {
		for (unsigned j = 0; j < cwins[i]->ins.size(); j++) {
			int k = cwins[i]->ins[j];
			if (k >= (int)cwins.size()) {
				continue; // primary input
			}
			CWinEdge *edge = new CWinEdge(cwins[k], cwins[i]); // k -> i
			cwins[i]->edges.add(edge);
			cwins[k]->edges.add(edge);
			edgeN++;
		}
	}
}

void Circuit::setGateInN() {
	gateInN = 0;
	vector<string> v;
	for (unsigned i = 0; i < cwins.size(); i++) {
		parseNames(cwins[i]->getGates()->stor[0], v);
		gateInN += v.size() - 2; // except the keyword and output
	}
}

void Circuit::setLevN() {
	levN = 0;
	int n = cwins.size();
	vector<int> levels, gatesIns; // gate levels, primary input gates
	levels.reserve(n);
	gatesIns.reserve(n);
	for (int i = 0; i < n; i++) {
		levels.push_back(LEVEL_FRESH);
		for (unsigned j = 0; j < cwins[i]->ins.size(); j++) {
			if (cwins[i]->ins[j] >= (int)cwins.size()) {
				gatesIns.push_back(i);
				break;
			}
		}
	}
	vector<LevItem> stack;
	stack.reserve(n);
	for (unsigned i = 0; i < gatesIns.size(); i++) {
		LevItem item;
		item.cwinI = gatesIns[i];
		item.edgeI = 0;
		stack.push_back(item);
		while (stack.size()) {
			int
				cwinI = stack.back().cwinI,
				&edgeI = stack.back().edgeI,
				&level = levels[cwinI];
			if (level == LEVEL_FRESH) {
				level = 0;
			}
			if (edgeI >= (int)cwins[cwinI]->edges.size()) {
				if (level) {
					level++;
				}
				else if (outs.find(cwins[cwinI]->outs[0])) {
					level = 1;
				}
				stack.pop_back();
				continue;
			}
			if (cwins[cwinI] != cwins[cwinI]->edges[edgeI]->getCWinA()) {
					// not an out edge
				edgeI++;
				continue;
			}
			item.cwinI = cwins[cwinI]->edges[edgeI]->getCWinB()->getId();
			int level2 = levels[item.cwinI];
			if (level2 == LEVEL_FRESH) {
				stack.push_back(item);
				continue;
			}
			if (level2 > level) {
				level = level2;
			}
			edgeI++;
		}
		int val = levels[gatesIns[i]];
		if (val > levN) {
			levN = val;
		}
	}
	if (levN < 1) {
		logErr(ERR_BLIF, "Invalid netlist in file " + file);
	}
}

void Circuit::buildEdgeHeap(Heap &heap, const Params &pm, TrieStor<CWin *> &neighbors,
	TrieStor<int> &workIns, TrieStor<int> &workOuts, TrieStor<int> &workIsect
) const {
	heap.reserve(edgeN);
	heap.setValid(false); // faster performance
	for (unsigned i = 0; i < cwins.size(); i++) {
		for (unsigned j = 0; j < cwins[i]->edges.size(); j++) {
			CWinEdge *edge = cwins[i]->edges[j];
			if (cwins[i] == edge->getCWinA()) { // no duplicities
				edge->setEval(pm, outs, neighbors, workIns, workOuts, workIsect);
				heap.insert(edge);
			}
		}
	}
	heap.setValid(true); // build
}

void Circuit::removeEdges(CWin *cwin, Heap &heap) {
	for (unsigned i = 0; i < cwin->edges.size(); i++) {
		heap.remove(cwin->edges[i]);
	}
	cwin->removeEdges();
}

void Circuit::addEdges(CWin *cwin, Heap &heap, const Params &pm, TrieStor<CWin *> &neighbors,
	TrieStor<int> &workIns, TrieStor<int> &workOuts, TrieStor<int> &workIsect
) const {
	for (unsigned i = 0; i < neighbors.size(); i++) {
		CWinEdge *edge = new CWinEdge(cwin, neighbors[i]);
		cwin->edges.add(edge);
		neighbors[i]->edges.add(edge);
	}
	for (unsigned i = 0; i < cwin->edges.size(); i++) {
		CWinEdge *edge = cwin->edges[i];
		edge->setEval(pm, outs, neighbors, workIns, workOuts, workIsect);
		heap.insert(edge);
	}
}
