#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <sstream>
#include <climits>

using namespace std;

#define NO_INDEX -1

string toString(int x) {
	stringstream ss;
	ss << x;
	return ss.str();
}

template<typename T>
class TrieStor {
private:
	vector<int> trie;
	vector<T> data;
private: // specializations
	void reset_0();
	int getPos(const T &) const;
	int getPosA(const T &); // appends the val into trie
public:
	TrieStor();
	void reset();
	void reserveTrie(int);
	void reserveData(int);
	const vector<int> &getTrie() const;
	const vector<T> &getData() const;
	unsigned size() const;
	bool add(const T &);
	bool remove(const T &);
	int getI(const T &) const;
	bool find(const T &) const;
	const T &operator[](int i) const;
	TrieStor<T> &operator=(const TrieStor<T> &x);
	TrieStor<T> &operator+=(const TrieStor<T> &x);
	TrieStor<T> &operator-=(const TrieStor<T> &x);
	bool isect(const TrieStor<T> &x, const TrieStor<T> &y);
private:
	void reset_str();
	void newNode();
	int getPos_key(const char *, int n) const;
	int getPosA_key(const char *, int n);
	int getPos_str(const string &) const;
	int getPosA_str(const string &);
};

template<typename T>
void TrieStor<T>::reset_0() {
}

template<>
void TrieStor<string>::reset_0() {
	reset_str();
}

template<typename T>
int TrieStor<T>::getPos(const T &x) const {
	return getPos_key((const char *)&x, sizeof(T));
}

template<>
int TrieStor<string>::getPos(const string &x) const {
	return getPos_str(x);
}

template<typename T>
int TrieStor<T>::getPosA(const T &x) {
	return getPosA_key((const char *)&x, sizeof(T));
}

template<>
int TrieStor<string>::getPosA(const string &x) {
	return getPosA_str(x);
}

template<typename T>
TrieStor<T>::TrieStor() {
	reset();
}

template<typename T>
void TrieStor<T>::reset() {
	data.clear();
	trie.clear();
	newNode(); // pushing the root
	reset_0(); // type specific tail
}

template<typename T>
void TrieStor<T>::reserveTrie(int n) {
	trie.reserve(n);
}

template<typename T>
void TrieStor<T>::reserveData(int n) {
	data.reserve(n);
}

template<typename T>
const vector<int> &TrieStor<T>::getTrie() const {
	return trie;
}

template<typename T>
const vector<T> &TrieStor<T>::getData() const {
	return data;
}

template<typename T>
unsigned TrieStor<T>::size() const {
	return data.size();
}

template<typename T>
bool TrieStor<T>::add(const T &x) {
	int pos = getPosA(x);
	if (trie[pos] == NO_INDEX) {
		trie[pos] = data.size();
		data.push_back(x);
		return true;
	}
	return false;
}

template<typename T>
bool TrieStor<T>::remove(const T &x) {
	int
		pos = getPos(x),
		itemI = (pos != NO_INDEX ? trie[pos] : NO_INDEX);
	if (itemI == NO_INDEX) {
		return false;
	}
	if (itemI != (int)data.size() - 1) {
		trie[getPos(data.back())] = itemI;
		data[itemI] = data.back();
	}
	trie[pos] = NO_INDEX;
	data.pop_back();
	return true;
}

template<typename T>
int TrieStor<T>::getI(const T &x) const {
	int pos = getPos(x);
	return (pos != NO_INDEX ? trie[pos] : NO_INDEX);
}

template<typename T>
bool TrieStor<T>::find(const T &x) const {
	return (getI(x) != NO_INDEX);
}

template<typename T>
const T &TrieStor<T>::operator[](int i) const {
	return data[i];
}

template<typename T>
TrieStor<T> &TrieStor<T>::operator=(const TrieStor<T> &x) {
	if (&x == this) {
		return *this;
	}
	reset();
	return (*this += x);
}

template<typename T>
TrieStor<T> &TrieStor<T>::operator+=(const TrieStor<T> &x) {
	if (&x == this) {
		return *this;
	}
	data.reserve(x.data.size());
	for (unsigned i = 0; i < x.data.size(); i++) {
		add(x.data[i]);
	}
	return *this;
}

template<typename T>
TrieStor<T> &TrieStor<T>::operator-=(const TrieStor<T> &x) {
	if (&x == this) {
		reset();
		return *this;
	}
	for (unsigned i = 0; i < x.data.size(); i++) {
		remove(x.data[i]);
	}
	return *this;
}

template<typename T>
bool TrieStor<T>::isect(const TrieStor<T> &x, const TrieStor<T> &y) {
	if (&x == this || &y == this) {
		return false;
	}
	bool by = (y.data.size() > x.data.size());
	const TrieStor<T>
		*bp = (by ? &y : &x),
		*sp = (by ? &x : &y); // smaller one, used in cycle because of O(1) iterations
	reset();
	for (unsigned i = 0; i < sp->data.size(); i++) {
		if (bp->getI(sp->data[i]) != NO_INDEX) {
			add(sp->data[i]);
		}
	}
	return true;
}

template<typename T>
void TrieStor<T>::reset_str() {
	trie.push_back(NO_INDEX); // for empty string
}

template<typename T>
void TrieStor<T>::newNode() {
	trie.push_back(NO_INDEX); // for 0
	trie.push_back(NO_INDEX); // for 1
}

template<typename T>
int TrieStor<T>::getPos_key(const char *x, int n) const {
	int pos = 0;
	for (int i = n - 1; i >= 0; i--) {
		char byte = x[i];
		for (int j = CHAR_BIT - 1; j >= 0; j--) {
			char bit = (byte >> j) & 1;
			if (trie[pos + bit] != NO_INDEX) {
				pos = trie[pos + bit];
			}
			else {
				return NO_INDEX;
			}
		}
	}
	return pos;
}

template<typename T>
int TrieStor<T>::getPosA_key(const char *x, int n) {
	int pos = 0;
	for (int i = n - 1; i >= 0; i--) {
		char byte = x[i];
		for (int j = CHAR_BIT - 1; j >= 0; j--) {
			char bit = (byte >> j) & 1;
			if (trie[pos + bit] != NO_INDEX) {
				pos = trie[pos + bit];
			}
			else {
				trie[pos + bit] = trie.size();
				pos = trie.size();
				if (i > 0 || j > 0) {
					newNode();
				}
				else {
					trie.push_back(NO_INDEX); // item index
				}
			}
		}
	}
	return pos;
}

template<typename T>
int TrieStor<T>::getPos_str(const string &x) const {
	int pos = 0;
	for (unsigned i = 0; i < x.size(); i++) {
		char byte = x[i];
		for (int j = CHAR_BIT - 1; j >= 0; j--) {
			char bit = (byte >> j) & 1;
			if (trie[pos + bit] != NO_INDEX) {
				pos = trie[pos + bit];
			}
			else {
				return NO_INDEX;
			}
		}
	}
	return pos + 2;
}

template<typename T>
int TrieStor<T>::getPosA_str(const string &x) {
	int pos = 0;
	for (unsigned i = 0; i < x.size(); i++) {
		char byte = x[i];
		for (int j = CHAR_BIT - 1; j >= 0; j--) {
			char bit = (byte >> j) & 1;
			if (trie[pos + bit] != NO_INDEX) {
				pos = trie[pos + bit];
			}
			else {
				trie[pos + bit] = trie.size();
				pos = trie.size();
				newNode();
				if (j == 0) {
					trie.push_back(NO_INDEX); // item index
				}
			}
		}
	}
	return pos + 2;
}

template<typename T>
void printCoef(const TrieStor<T> &stor) {
	int
		dataN = stor.size(),
		trieN = stor.getTrie().size();
	cout << dataN << ", " << trieN << ", " << (trieN / (double)dataN) << endl;
}

template<typename T>
void printData(const TrieStor<T> &stor) {
	cout << "data =";
	for (unsigned i = 0; i < stor.size(); i++) {
		cout << " " << stor[i];
	}
	cout << endl;
}

template<typename T>
void printTrie(const TrieStor<T> &stor) {
	const vector<int> &trie = stor.getTrie();
	for (unsigned i = 0; i < trie.size(); i++) {
		cout << " [" << i << ":" << trie[i] << "]";
	}
	cout << endl;
}

void test1(int n, int m) {
	TrieStor<string> stor;
	string x;
	string abc = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int count = 0;
	for (unsigned j = 0; j < abc.size(); j++) {
		for (int i = 0; i < n; i++) {
			x = abc[j];
			x += toString(i);
			if (!stor.add(x)) {
				cout << "error #1" << endl;
				exit(1);
			}
			if (++count % m == 0) {
				printCoef<string>(stor);
			}
		}
	}
	count = 0;
	for (unsigned j = 0; j < abc.size(); j++) {
		for (int i = 0; i < n; i++) {
			x = abc[j];
			x += toString(i);
			if (stor.getI(x) != count++) {
				cout << "error #2" << endl;
				exit(1);
			}
		}
	}
	for (unsigned j = 0; j < abc.size(); j++) {
		for (int i = 0; i < n; i++) {
			x = abc[j];
			x += toString(i);
			if (!stor.remove(x)) {
				cout << "error #3" << endl;
				exit(1);
			}
		}
	}
}

void test2(int n, int m) {
	TrieStor<int> stor;
	int count = 0;
	for (int i = 0; i < n; i++) {
		if (!stor.add(i)) {
			cout << "error #1" << endl;
			exit(1);
		}
		if (++count % m == 0) {
			printCoef<int>(stor);
		}
	}
	printData<int>(stor);
	count = 0;
	for (int i = 0; i < n; i++) {
		if (stor.getI(i) != count++) {
			cout << "error #2" << endl;
			exit(1);
		}
	}
	for (int i = 0; i < n; i++) {
		if (!stor.remove(i)) {
			cout << "error #3" << endl;
			exit(1);
		}
	}
}

void test3(char n) {
	TrieStor<char> stor;
	for (char i = 0; i < n; i++) {
		stor.add(i);
		printTrie<char>(stor);
	}
	cout << "---" << endl;
	for (char i = 0; i < n; i++) {
		stor.remove(i);
		printTrie<char>(stor);
	}
}

void test4() {
	TrieStor<char> a, b, c;
	for (char i = 0; i <= 3; i++) {
		a.add(i);
	}
	for (char i = 2; i <= 5; i++) {
		b.add(i);
	}
	c.isect(a, b);
	printTrie<char>(c);
	printTrie<char>(a);
	a += b;
	printTrie<char>(a);
	a -= b;
	printTrie<char>(a);
	a = b;
	printTrie<char>(a);
	a = a;
	printTrie<char>(a);
	a += a;
	printTrie<char>(a);
	a -= a;
	printTrie<char>(a);
}

int main() {
	cout << "hula" << endl;
	test2(20, 1);
	cout << "hula" << endl;
	return 0;
}
